#pragma once

#include <glm\glm.hpp>
#include <glm\gtc\matrix_transform.hpp>

#include <GLFW\glfw3.h>
#include <vector>

class Camera
{
public:
	Camera();
	Camera(glm::vec3 startPosition, glm::vec3 startUp, GLfloat startYaw, GLfloat startPitch, GLfloat startMoveSpeed, GLfloat startTurnSpeed);

	void keyControl(bool* keys, GLfloat deltaTime);
	void mouseControl(GLfloat xChange, GLfloat yChange);

	glm::vec3 getCameraPosition();
	glm::vec3 getCameraDirection();

	glm::mat4 calculateViewMatrix();

	void setPosition(glm::vec3 pos);
	void setFront(glm::vec3 front);
	void setTotalTime(GLfloat time);

	void setControlPoints(const std::vector<glm::vec3>& controlPoints);
	bool followPathLookAhead(GLfloat deltaTime);
	bool followPathLookDirection(GLfloat deltaTime, const glm::vec3* direction);
	bool followPathLookTarget(GLfloat deltaTime, const glm::vec3* targetPosition);
	glm::vec3 getTargetPositionOnSpline(float t, const std::vector<glm::vec3>& targetControlPoints);

	void update();
	~Camera();

private:
	glm::vec3 position;
	glm::vec3 front;
	glm::vec3 up;
	glm::vec3 right;
	glm::vec3 worldUp;

	GLfloat yaw;
	GLfloat pitch;

	GLfloat moveSpeed;
	GLfloat turnSpeed;

	bool demoMode;

	std::vector<glm::vec3> controlPoints;
	GLfloat totalTime;
	glm::vec3 targetPosition;
	float pathTime;
	glm::vec3 getPositionOnSpline(float t);
};

