#include "ConcertLight.h"

Model dummyModel;
SpotLight dummySpotlight;

ConcertLight::ConcertLight() :
    lamp(dummyModel),
    spotlight(dummySpotlight),
    spotlightOffset(glm::vec3(0.0f, 1.0f, 0.0f)),
    position(0.0f),
    rotation(0.0f)
{}

ConcertLight::ConcertLight(Model lamp,
    SpotLight spotlight,
    glm::vec3 spotlightOffset) :
    lamp(lamp),
    spotlight(spotlight),
    spotlightOffset(spotlightOffset)
{}

void ConcertLight::updateSpotlight() {
    
    glm::vec3 spotlightDirection = glm::vec3(0.0f, -1.0f, 0.0f);

    spotlight.setPosition(spotlightDirection);
    spotlight.setPosition(spotlightDirection);
}

glm::vec3 ConcertLight::getPosition() const {
    return position;
}

glm::vec3 ConcertLight::getRotation() const {
    return rotation;
}

void ConcertLight::SetPosition(glm::vec3 newPos) {
    position = newPos;
}

void ConcertLight::SetRotation(glm::vec3 newRot) {
    rotation = newRot;
}

void ConcertLight::lookDown() {
    rotation = glm::vec3(0.0, -1.0, 0.0);

}

void ConcertLight::lookStage() {
    glm::vec3 target = glm::vec3(-0.0165168f, -6.2008f, 17.6032f); // Stage position
    glm::vec3 direction = glm::normalize(target - position);
    rotation = direction;
}

void ConcertLight::lookDJ() {
    glm::vec3 target = glm::vec3(0.725199f, 5.30762f, 3.36215f); // DJ position
    glm::vec3 direction = glm::normalize(target - position);
    rotation = direction;
}

void ConcertLight::blink(bool on, GLfloat deltaTime) {
    
    static GLfloat time = 0.0f;
    static bool isOn = true; // Start with the light on

    // Calculate seconds per beat
    GLfloat bpm = 157.0f;
    GLfloat secondsPerBeat = (60.0f / bpm) *2;

    time += deltaTime;

    if (time >= secondsPerBeat) {
        time = 0.0f; // Reset time
        isOn = !isOn; // Toggle the light state
    }

    if (on) {
        if (isOn) {
            spotlight.setIntensity(10.0f); // Set to max intensity when on
        }
        else {
            spotlight.setIntensity(0.5f); // Set to min intensity when off
        }
    }
    
}

void ConcertLight::swirl() {
    static float time = 0.0f;
    static bool isSwirlingOut = true;

    time += 0.002f;

    float angle = time * 360.0f;
    if (!isSwirlingOut) {
        angle = -angle;
    }

    glm::vec3 swirlRotation = glm::vec3(position.x > 0 ? angle : -angle, -90.0f, 90.0f);
    SetRotation(swirlRotation);

    if (time >= 3.0f) {
        time = 0.0f;
        isSwirlingOut = !isSwirlingOut;
    }
}

void ConcertLight::swirlButFaster(GLfloat deltaTime) {
    static float time = 0.0f;
    static int currentRotationIndex = 0;
    static glm::vec3 initialRotation = rotation; // Store the initial rotation

    time += deltaTime;

    // Define the rotation sequence
    std::vector<glm::vec3> rotations = {
        initialRotation,              // Starting position
        glm::vec3(-glm::sign(position.x), -1.0f, 0.0f),
        glm::vec3(0.0f, 0.0f, 1.0f), 
        glm::vec3(glm::sign(position.x), 1.0f, 0.0f),
        glm::vec3(0.0f, 0.0f, -1.0f),
        initialRotation
    };

    // Calculate the interpolation parameter (t)
    float t = time / 16.0f;

    // Calculate the current rotation index
    currentRotationIndex = static_cast<int>(t * rotations.size()) % rotations.size();

    // Interpolate between the current and next rotation
    int nextRotationIndex = (currentRotationIndex + 1) % rotations.size();
    glm::vec3 interpolatedRotation = glm::mix(rotations[currentRotationIndex], rotations[nextRotationIndex], t - static_cast<int>(t));

    SetRotation(interpolatedRotation);

    if (time >= 16.0f) {
        time = 0.0f;
        currentRotationIndex = 0; // Reset to the initial rotation
    }
}

void ConcertLight::colorChange(glm::vec3 color) {
    spotlight.setColor(color);
}

Model& ConcertLight::getLamp() {
    return lamp;
}

SpotLight& ConcertLight::getSpotlight() {
    return spotlight;
}