// DemoController.h

#ifndef DEMO_CONTROLLER_H
#define DEMO_CONTROLLER_H

#include <glm/glm.hpp>
#include "Camera.h"
#include "CameraMovement.h"
#include <queue>
#include <irrKlang.h>
#include <chrono>

class Camera;

class DemoController {
public:
    DemoController(Camera& camera, irrklang::ISound* playingSound);
    bool isDemoMode() const { return demoMode; }
    void startDemo(Camera& camera);
    void moveCamera(GLfloat deltaTime);
    void endDemo();
    void processMovementQueue(GLfloat deltaTime);

private:
    Camera& camera;
    irrklang::ISound* playingSound;
    bool demoMode;
    glm::vec3 cameraPosAtStart;
    glm::vec3 cameraDirAtStart;
    std::chrono::system_clock::time_point startTime;
    std::queue<CameraMovement*> movementQueue;
    void defineControllPoints();
    std::vector<glm::vec3> controlPoints;
    CameraMovement* currentMovement;
    std::chrono::system_clock::time_point startMovementTime;
    std::chrono::system_clock::time_point endMovementTime;
    std::chrono::system_clock::time_point currentTime;
    void printTimePoint(const std::chrono::system_clock::time_point& timePoint);
    GLfloat beatsToSeconds(GLfloat numBeats);
};

#endif 