#include "ParticleLinearMovement.h"
#include <glm/glm.hpp>
#include <glm/gtx/compatibility.hpp> // For glm::mix
#include <cmath>                    // For fmod

ParticleLinearMovement::ParticleLinearMovement(glm::vec3 startPoint, glm::vec3 endPoint, float duration)
    : startPoint(startPoint), endPoint(endPoint), duration(duration), startTime(0.0f) {}

glm::vec3 ParticleLinearMovement::GetPosition(float currentTime) {
    // Calculate the total time elapsed since the movement started
    float elapsedTime = currentTime - startTime;

    // Wrap elapsedTime to keep it within [0, duration]
    float wrappedTime = fmod(elapsedTime, duration);

    // Normalize time to [0.0, 1.0] for interpolation
    float t = wrappedTime / duration;

    // Interpolate position between startPoint and endPoint
    return glm::mix(startPoint, endPoint, t);
}

void ParticleLinearMovement::Start(float currentTime) {
    startTime = currentTime;
}
