#ifndef PARTICLESYSTEM_H
#define PARTICLESYSTEM_H

#include <GL/glew.h>
#include <glm/glm.hpp>
#include <glm/gtc/type_ptr.hpp>
#include <GLFW\glfw3.h>
#include <vector>
#include <string>
#include <iostream>
#include <fstream>
#include <sstream>
#include <ctime>
#include "IParticleMovement.h"

class ParticleSystem {
public:
    ParticleSystem(GLuint particleCount, const float TTL, float particleSize, double spawnRatePerSecond, glm::vec3 baseColor, float nearZ, float farZ);
    ~ParticleSystem();

    void Initialize();
    void Update(float deltaTime);
    void SetMovementPattern(IParticleMovement* newPattern);
    void Render(const glm::mat4& viewMatrix, const glm::mat4& projectionMatrix, const glm::vec3& cameraPosition, const GLuint depthTexture, const glm::vec2 screenDimension);

private:
    const float nearZ;
    const float farZ;
    GLuint particleCount;
    const float TTL;
    float particleSize;
    const double spawnRatePerSecond;
    glm::vec3 baseColor;

    IParticleMovement* movementPattern;  // Delegates movement logic
    glm::vec3 newParticlePosition;       // Position from movement logic

    GLuint posSSBO[2], velSSBO[2];
    int index;
    GLuint atomicCounter;
    GLuint computeShaderProgram;
    GLuint renderProgram;
    GLuint particle_vaos[2];
    const GLuint particle_position_location = 0;

    std::string LoadShaderCode(const char* filePath);
    GLuint CompileShader(const char* filePath, GLenum shaderType);
    GLuint LinkProgram(GLuint vertexShader, GLuint geometryShader, GLuint fragmentShader);
    void InitializeParticleBuffers();
    void InitializeAtomicCounter();
    void CreateComputeShader();
    void InitializeParticleRendering();
    void AddInitialParticles();
    void enableBlendMode();
    void disableBlendMode();
};

#endif // PARTICLESYSTEM_H
