#version 460

in vec2 TexCoord0;
flat in float TTL0;
uniform vec3 BaseColor;
uniform sampler2D depthTexture;  // Depth texture from G-buffer
uniform mat4 ProjectionMatrix;
uniform float zNear;
uniform float zFar;
uniform vec2 screenDimensions;   // Screen dimensions

out vec4 FragColor;

// Function to linearize depth value
float LinearizeDepth(float depth) {
    float z = depth * 2.0 - 1.0; // Back to NDC 
    return (2.0 * zNear * zFar) / (zFar + zNear - z * (zFar - zNear));    
}

vec3 pow3(vec3 v, float p) { 
    return pow(abs(v), vec3(p));
}

vec3 glowSphere(vec3 color, vec2 p, float radius, float glow) {
    float term = 1.0 / length(p) / radius;
    return pow3(color * term * 0.5, glow) * radius;
}

void main() {
    
    // Calculate depth in clip space using projection matrix
    float particleDepth = LinearizeDepth(gl_FragCoord.z); // Linearize particle depth
    // Fetch depth from G-buffer and linearize
    float sceneDepth = texture(depthTexture, gl_FragCoord.xy / screenDimensions).r;
    sceneDepth = LinearizeDepth(sceneDepth);
    // Perform depth test
    if (particleDepth > sceneDepth) {
        discard; // Particle is occluded
    }
    
    // Interpolate from red to black based on TTL
    vec3 baseColor = mix(vec3(0.0, 0.0, 0.0), BaseColor, clamp(TTL0, 0.0, 1.0));
    
    vec3 color = glowSphere(baseColor, TexCoord0 * 2.0 - 1.0, 5.0, 3.25);
    float lum = dot(color, vec3(0.299, 0.587, 0.114));

    

    FragColor = vec4(color, lum); // Use luminance as alpha
}
