#version 330 core
out float FragColor;

in vec2 TexCoords;

uniform sampler2D gPosition;  // View-space positions
uniform sampler2D gNormal;    // View-space normals
uniform sampler2D noiseTexture;

uniform vec3 samples[64];
uniform mat4 projection;      // Projection matrix for view to clip space

const vec2 noiseScale = vec2(800.0 / 4.0, 600.0 / 4.0);
const int kernelSize = 64;
const float radius = 1.0;
const float bias = 0.01; 


void main() {
    vec3 fragPosView = texture(gPosition, TexCoords).xyz; // Assume view space
    vec3 normalView = texture(gNormal, TexCoords).rgb;    // Assume view space
    vec3 randomVec = texture(noiseTexture, TexCoords * noiseScale).xyz;

    // Compute tangent and bitangent vectors to create TBN matrix
    vec3 tangent = normalize(randomVec - normalView * dot(randomVec, normalView));
    vec3 bitangent = cross(normalView, tangent);
    mat3 TBN = mat3(tangent, bitangent, normalView);

    float occlusion = 0.0;
    for (int i = 0; i < kernelSize; ++i) {
        // Transform sample from tangent space to view space
        vec3 sampleVec = TBN * samples[i];
        vec3 samplePos = fragPosView + sampleVec * radius;

        // Transform samplePos to screen space before sampling depth buffer
        vec4 offset = projection * vec4(samplePos, 1.0);
        offset.xyz /= offset.w;  // Perspective divide to NDC
        offset.xyz = offset.xyz * 0.5 + 0.5; // NDC to [0, 1] range

        float sampleDepth = texture(gPosition, offset.xy).z;
        float rangeCheck = smoothstep(0.0, 1.0, radius / abs(fragPosView.z - sampleDepth));
        occlusion += (sampleDepth >= samplePos.z + bias ? 1.0 : 0.0) * rangeCheck;
    }

    occlusion = 1.0 - (occlusion / kernelSize);
    FragColor = occlusion;
}
