#define STB_IMAGE_IMPLEMENTATION

#include <stdio.h>
#include <string.h>
#include <cmath>
#include <vector>
#include <sstream>
#include <iostream>
#include <fstream>
#include <random>

#include <GL/glew.h>
#include <GLFW/glfw3.h>

#include <glm/glm.hpp>
#include <glm/gtc/matrix_transform.hpp>
#include <glm/gtc/type_ptr.hpp>
#include <glm/gtx/string_cast.hpp>
#include <glm/gtx/quaternion.hpp> 
#include <glm/gtx/vector_angle.hpp>

#include "CommonValues.h"
#include "Window.h"
#include "Mesh.h"
#include "Camera.h"
#include "Texture.h"
#include "DirectionalLight.h"
#include "PointLight.h"
#include "SpotLight.h"
#include "Material.h"
#include "Model.h"
#include "DemoController.h"
#include "ConcertLight.h"

#include "ParticleSystem.h"
#include "IParticleMovement.h"       // Movement interface
#include "ParticleCircularMovement.h" // Circular movement implementation
#include "ParticleLinearMovement.h"   // Linear movement implementation
#include <assimp/Importer.hpp>
#include <irrKlang.h>

using namespace irrklang;


// Window and Camera variables
Window mainWindow;
Camera camera;
const float nearZ = 0.1f;
const float farZ = 100.0f;
glm::mat4 projection;

// Models
Model tree;
Model terrain;
Model stage;
Model robotT; // robot Torso
Model robotRL; // robot Right Leg
Model robotLL; // robot Left Leg
Model robotRLL; // robot Right Lower Leg
Model robotLLL; // robot Left Lower Leg
Model robotRA; // robot Right Arm
Model robotLA; // robot Left Arm
Model robotH; // robot Head
Model lamp;
// Model lamps[MAX_SPOT_LIGHTS]; // size is same as spotlights

Model DJT; // DJ Torso
Model DJRL; // DJ Right Leg
Model DJLL; // DJ Left Leg
Model DJRA; // DJ Right Arm
Model DJLA; // DJ Left Arm
Model DJH; // DJ Head

Model Mixtable; // DJ Mixtable



glm::vec3 orangeColor = glm::vec3(239.0f / 255.0f, 144.0f / 255.0f, 6.0f / 255.0f);
glm::vec3 purpleColor = glm::vec3(143.0f / 255.0f, 63.0f / 255.0f, 166.0f / 255.0f);

// Particle Systems
ParticleSystem* circularParticleSystemStage;
ParticleSystem* circularParticleSystemStage2;
ParticleSystem* linearParticleSystem;
ParticleSystem* linearParticleSystem2;
ParticleSystem* linearParticleSystem3;
ParticleSystem* linearParticleSystem4;
ParticleSystem* linearParticleSystem5;
ParticleSystem* linearParticleSystem6;
ParticleSystem* linearParticleSystem7;
ParticleSystem* linearParticleSystem8;
ParticleSystem* linearParticleSystem9;
ParticleSystem* linearParticleSystem10;
ParticleSystem* linearParticleSystem11;
ParticleSystem* linearParticleSystem12;

// Lights
DirectionalLight mainDirectionalLight;
PointLight pointLights[MAX_POINT_LIGHTS];
// SpotLight spotLights[MAX_SPOT_LIGHTS];

// ConcertLights
ConcertLight concertLights[MAX_SPOT_LIGHTS];
ConcertLight singleLight;


// Deferred Shading Buffers
GLuint gBuffer;
GLuint gPosition, gNormal, gAlbedoSpec;
GLuint rboDepth;
// Deferred Shading Textures
GLuint depthTexture;
GLuint lightingTexture;
GLuint lightingFramebuffer;

// SSAO variables
GLuint ssaoFBO, ssaoBlurFBO;
GLuint ssaoColorBuffer, ssaoColorBufferBlur;
GLuint noiseTexture;
std::vector<glm::vec3> ssaoKernel;

// Display Texture Variables
GLuint quadVAO = 0, quadVBO;

// Shaders
GLuint geometryShader;
GLuint lightingShader;
GLuint combineShader;
GLuint displayShader;
GLuint ssaoShader;
GLuint ssaoBlurShader;

// Variable to decide the rendermode
int displayMode = 0;

// Time Variables
GLfloat deltaTime = 0.0f;

// Demo Lights variables
float demoStartTime;
bool swirling = false;

// Define the Starting Points for all Lamps and Spotlights
std::vector<std::pair<glm::vec3, glm::vec3>> lampTransforms = {
	{glm::vec3(-7.38, 10.6, 8.4), glm::vec3(0.0f, -1.0f, 0.0f)},
	{glm::vec3(-4.45, 10.6, 8.4), glm::vec3(0.0f, -1.0f, 0.0f)},
	{glm::vec3(-1.47, 10.6, 8.4), glm::vec3(0.0f, -1.0f, 0.0f)},
	{glm::vec3(1.39, 10.6, 8.4), glm::vec3(0.0f, -1.0f, 0.0f)},
	{glm::vec3(4.44, 10.6, 8.4), glm::vec3(0.0f, -1.0f, 0.0f)},
	{glm::vec3(7.33, 10.6, 8.4), glm::vec3(0.0f, -1.0f, 0.0f)},

	{glm::vec3(-12.0f, 10.6f, 10.0f), glm::vec3(0.0f, -1.0f, 0.0f)},
	{glm::vec3(-12.0f, 10.6f, 13.0f), glm::vec3(0.0f, -1.0f, 0.0f)},
	{glm::vec3(-12.0f, 10.6f, 16.0f), glm::vec3(0.0f, -1.0f, 0.0f)},
	{glm::vec3(-12.0f, 10.6f, 19.0f), glm::vec3(0.0f, -1.0f, 0.0f)},

	{glm::vec3(12.0f, 10.6f, 10.0f), glm::vec3(0.0f, -1.0f, 0.0f)},
	{glm::vec3(12.0f, 10.6f, 13.0f), glm::vec3(0.0f, -1.0f, 0.0f)},
	{glm::vec3(12.0f, 10.6f, 16.0f), glm::vec3(0.0f, -1.0f, 0.0f)},
	{glm::vec3(12.0f, 10.6f, 19.0f), glm::vec3(0.0f, -1.0f, 0.0f)}
};

// Helper function for linear interpolation
float lerp(float a, float b, float f) {
	return a + f * (b - a);
}
// Function to load shaders from file
GLuint LoadShaders(const char* vertex_file_path, const char* fragment_file_path) {
	// Create the shaders
	GLuint VertexShaderID = glCreateShader(GL_VERTEX_SHADER);
	GLuint FragmentShaderID = glCreateShader(GL_FRAGMENT_SHADER);

	// Read the Vertex Shader code from the file
	std::string VertexShaderCode;
	std::ifstream VertexShaderStream(vertex_file_path, std::ios::in);
	if (VertexShaderStream.is_open()) {
		std::stringstream sstr;
		sstr << VertexShaderStream.rdbuf();
		VertexShaderCode = sstr.str();
		VertexShaderStream.close();
	}
	else {
		printf("Impossible to open %s. Are you in the right directory?\n", vertex_file_path);
		getchar();
		return 0;
	}

	// Read the Fragment Shader code from the file
	std::string FragmentShaderCode;
	std::ifstream FragmentShaderStream(fragment_file_path, std::ios::in);
	if (FragmentShaderStream.is_open()) {
		std::stringstream sstr;
		sstr << FragmentShaderStream.rdbuf();
		FragmentShaderCode = sstr.str();
		FragmentShaderStream.close();
	}

	GLint Result = GL_FALSE;
	int InfoLogLength;

	// Compile Vertex Shader
	printf("Compiling shader: %s\n", vertex_file_path);
	char const* VertexSourcePointer = VertexShaderCode.c_str();
	glShaderSource(VertexShaderID, 1, &VertexSourcePointer, NULL);
	glCompileShader(VertexShaderID);

	// Check Vertex Shader
	glGetShaderiv(VertexShaderID, GL_COMPILE_STATUS, &Result);
	glGetShaderiv(VertexShaderID, GL_INFO_LOG_LENGTH, &InfoLogLength);
	if (InfoLogLength > 0) {
		std::vector<char> VertexShaderErrorMessage(InfoLogLength + 1);
		glGetShaderInfoLog(VertexShaderID, InfoLogLength, NULL, &VertexShaderErrorMessage[0]);
		printf("%s\n", &VertexShaderErrorMessage[0]);
	}

	// Compile Fragment Shader
	printf("Compiling shader: %s\n", fragment_file_path);
	char const* FragmentSourcePointer = FragmentShaderCode.c_str();
	glShaderSource(FragmentShaderID, 1, &FragmentSourcePointer, NULL);
	glCompileShader(FragmentShaderID);

	// Check Fragment Shader
	glGetShaderiv(FragmentShaderID, GL_COMPILE_STATUS, &Result);
	glGetShaderiv(FragmentShaderID, GL_INFO_LOG_LENGTH, &InfoLogLength);
	if (InfoLogLength > 0) {
		std::vector<char> FragmentShaderErrorMessage(InfoLogLength + 1);
		glGetShaderInfoLog(FragmentShaderID, InfoLogLength, NULL, &FragmentShaderErrorMessage[0]);
		printf("%s\n", &FragmentShaderErrorMessage[0]);
	}

	// Link the program
	printf("Linking program\n");
	GLuint ProgramID = glCreateProgram();
	glAttachShader(ProgramID, VertexShaderID);
	glAttachShader(ProgramID, FragmentShaderID);
	glLinkProgram(ProgramID);

	// Check the program
	glGetProgramiv(ProgramID, GL_LINK_STATUS, &Result);
	glGetProgramiv(ProgramID, GL_INFO_LOG_LENGTH, &InfoLogLength);
	if (InfoLogLength > 0) {
		std::vector<char> ProgramErrorMessage(InfoLogLength + 1);
		glGetProgramInfoLog(ProgramID, InfoLogLength, NULL, &ProgramErrorMessage[0]);
		printf("%s\n", &ProgramErrorMessage[0]);
	}

	glDetachShader(ProgramID, VertexShaderID);
	glDetachShader(ProgramID, FragmentShaderID);

	glDeleteShader(VertexShaderID);
	glDeleteShader(FragmentShaderID);

	return ProgramID;
}

void InitializeShaders() {
	geometryShader = LoadShaders("Shaders/deferred_geometry_vertex.glsl", "Shaders/deferred_geometry_fragment.glsl");
	lightingShader = LoadShaders("Shaders/deferred_lighting_vertex.glsl", "Shaders/deferred_lighting_fragment.glsl");
	combineShader = LoadShaders("Shaders/deferred_combine_vertex.glsl", "Shaders/deferred_combine_fragment.glsl");
	displayShader = LoadShaders("Shaders/display_vertex.glsl", "Shaders/display_fragment.glsl");
	ssaoShader = LoadShaders("Shaders/ssao_vertex.glsl", "Shaders/ssao_fragment.glsl");
	ssaoBlurShader = LoadShaders("Shaders/ssao_blur_vertex.glsl", "Shaders/ssao_blur_fragment.glsl");
}

// Function to set up G-Buffer for deferred shading
void setupGBuffer() {
	glGenFramebuffers(1, &gBuffer);
	glBindFramebuffer(GL_FRAMEBUFFER, gBuffer);

	// Position color buffer
	glGenTextures(1, &gPosition);
	glBindTexture(GL_TEXTURE_2D, gPosition);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA16F, mainWindow.getBufferWidth(), mainWindow.getBufferHeight(), 0, GL_RGBA, GL_FLOAT, NULL);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, gPosition, 0); //ATTACHMENT 0

	// Color + Specular color buffer
	glGenTextures(1, &gAlbedoSpec);
	glBindTexture(GL_TEXTURE_2D, gAlbedoSpec);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, mainWindow.getBufferWidth(), mainWindow.getBufferHeight(), 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT1, GL_TEXTURE_2D, gAlbedoSpec, 0);//ATTACHMENT 1

	// Normal color buffer
	glGenTextures(1, &gNormal);
	glBindTexture(GL_TEXTURE_2D, gNormal);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA16F, mainWindow.getBufferWidth(), mainWindow.getBufferHeight(), 0, GL_RGBA, GL_FLOAT, NULL);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT2, GL_TEXTURE_2D, gNormal, 0); //ATTACHMENT 2

	// Depth buffer texture
	glGenTextures(1, &depthTexture);
	glBindTexture(GL_TEXTURE_2D, depthTexture);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT, mainWindow.getBufferWidth(), mainWindow.getBufferHeight(), 0, GL_DEPTH_COMPONENT, GL_FLOAT, NULL);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glFramebufferTexture2D(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_TEXTURE_2D, depthTexture, 0);

	// Tell OpenGL which color attachments we'll use (of this framebuffer) for rendering
	GLuint attachments[3] = { GL_COLOR_ATTACHMENT0, GL_COLOR_ATTACHMENT1, GL_COLOR_ATTACHMENT2 };
	glDrawBuffers(3, attachments);

	// Finally check if framebuffer is complete
	if (glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE) {
		std::cout << "Framebuffer not complete!" << std::endl;
		switch (glCheckFramebufferStatus(GL_FRAMEBUFFER)) {
		case GL_FRAMEBUFFER_UNDEFINED:
			std::cout << "Framebuffer Undefined" << std::endl;
			break;
		case GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT:
			std::cout << "Framebuffer Incomplete Attachment" << std::endl;
			break;
		case GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT:
			std::cout << "Framebuffer Incomplete Missing Attachment" << std::endl;
			break;
		case GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER:
			std::cout << "Framebuffer Incomplete Draw Buffer" << std::endl;
			break;
		case GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER:
			std::cout << "Framebuffer Incomplete Read Buffer" << std::endl;
			break;
		case GL_FRAMEBUFFER_UNSUPPORTED:
			std::cout << "Framebuffer Unsupported" << std::endl;
			break;
		case GL_FRAMEBUFFER_INCOMPLETE_MULTISAMPLE:
			std::cout << "Framebuffer Incomplete Multisample" << std::endl;
			break;
		case GL_FRAMEBUFFER_INCOMPLETE_LAYER_TARGETS:
			std::cout << "Framebuffer Incomplete Layer Targets" << std::endl;
			break;
		default:
			std::cout << "Unknown Framebuffer Error" << std::endl;
			break;
		}
	}
	glBindFramebuffer(GL_FRAMEBUFFER, 0);
}

// Function to set up lighting framebuffer
void setupLightingFramebuffer() {
	glGenFramebuffers(1, &lightingFramebuffer);
	glBindFramebuffer(GL_FRAMEBUFFER, lightingFramebuffer);

	// Create texture to hold lighting pass result
	glGenTextures(1, &lightingTexture);
	glBindTexture(GL_TEXTURE_2D, lightingTexture);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA16F, mainWindow.getBufferWidth(), mainWindow.getBufferHeight(), 0, GL_RGBA, GL_FLOAT, NULL);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, lightingTexture, 0);

	// Depth buffer
	GLuint depthBuffer;
	glGenRenderbuffers(1, &depthBuffer);
	glBindRenderbuffer(GL_RENDERBUFFER, depthBuffer);
	glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT, mainWindow.getBufferWidth(), mainWindow.getBufferHeight());
	glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, depthBuffer);

	// Ensure framebuffer is complete
	if (glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE) {
		std::cout << "Framebuffer not complete!" << std::endl;
	}

	glBindFramebuffer(GL_FRAMEBUFFER, 0);
}


void LoadModels() {
	tree = Model();
	tree.LoadModel("Models/tree.obj");

	terrain = Model();
	terrain.LoadModel("Models/Floor/floor.obj");

	stage = Model();
	stage.LoadModel("Models/stage.obj");

	robotT = Model();
	robotT.LoadModel("Models/RobotParts/robotT.obj");
	robotRL = Model();
	robotRL.LoadModel("Models/RobotParts/robotRL.obj");
	robotLL = Model();
	robotLL.LoadModel("Models/RobotParts/robotLL.obj");
	robotRLL = Model();
	robotRLL.LoadModel("Models/RobotParts/robotRLL.obj");
	robotLLL = Model();
	robotLLL.LoadModel("Models/RobotParts/robotLLL.obj");
	robotRA = Model();
	robotRA.LoadModel("Models/RobotParts/robotRA.obj");
	robotLA = Model();
	robotLA.LoadModel("Models/RobotParts/robotLA.obj");
	robotH = Model();
	robotH.LoadModel("Models/RobotParts/robotH.obj");


	DJT = Model();
	DJT.LoadModel("Models/DJParts/DJT.obj");
	DJRL = Model();
	DJRL.LoadModel("Models/DJParts/DJRL.obj");
	DJLL = Model();
	DJLL.LoadModel("Models/DJParts/DJLL.obj");
	DJRA = Model();
	DJRA.LoadModel("Models/DJParts/DJRA.obj");
	DJLA = Model();
	DJLA.LoadModel("Models/DJParts/DJLA.obj");
	DJH = Model();
	DJH.LoadModel("Models/DJParts/DJH.obj");

	Mixtable = Model();
	Mixtable.LoadModel("Models/Mixtable/mixtable.obj");

	lamp = Model();
	lamp.LoadModel("Models/lamp.obj");
}

void initializeLights() {
	float pointAmbientIntensity = 1.0f;
	float pointDiffuseIntensity = 0.8f;
	
	// Initialize point lights
	pointLights[0] = PointLight(orangeColor.r,orangeColor.g,orangeColor.b, // orange
		pointAmbientIntensity, pointDiffuseIntensity,       // ambient and diffuse intensity
		0.0f, 1.0, 3.0f, // position
		1.0f, 0.09f, 0.032f); // attenuation constants
	


	float spotlightHeight = 8.0f;
	float spotAIntensity = 10.0f;
	float spotDIntensity = 1.0f;
	glm::vec3 spotDirection = glm::vec3(0.0f,-1.0f,-0.7f); 	

	Model currLamp;
	SpotLight currSpotLight;

	for (int i = 0; i < MAX_SPOT_LIGHTS; i++)
	{	
		currLamp = lamp;
		currSpotLight = SpotLight(1.0f, 1.0f, 1.0f,					// color
			spotAIntensity, spotDIntensity,                         // ambient and diffuse intensity
			lampTransforms[i].first.x, lampTransforms[i].first.y, lampTransforms[i].first.z,     // position
			lampTransforms[i].second.x, lampTransforms[i].second.y, lampTransforms[i].second.z,      // direction
			1.0f, 0.09f, 0.032f,									// constant, linear, exponential
			15.0f); // edge (cutoff angle in degrees)

		
		// Create the ConcertLight object
		concertLights[i] = ConcertLight(
			currLamp,
			currSpotLight,
			glm::vec3(0.0f, 0.0f, 0.0f) // Spotlight offset
		);

		concertLights[i].SetPosition(lampTransforms[i].first);
		concertLights[i].SetRotation(lampTransforms[i].second);

	}

	// Initialize directional light
	
	mainDirectionalLight = DirectionalLight(
		1.0f, 1.0f, 1.0f,    // color (white)
		0.3f,                // ambient intensity
		0.1f,                // diffuse intensity
		-0.2f, -1.0f, -0.3f  // direction          
	);
	
}

// Function to render a fullscreen quad
void renderQuad() {
	if (quadVAO == 0) {
		GLfloat quadVertices[] = {
			// Positions        // Texture Coords
			-1.0f,  1.0f, 0.0f, 0.0f, 1.0f,
			-1.0f, -1.0f, 0.0f, 0.0f, 0.0f,
			 1.0f, -1.0f, 0.0f, 1.0f, 0.0f,
			-1.0f,  1.0f, 0.0f, 0.0f, 1.0f,
			 1.0f, -1.0f, 0.0f, 1.0f, 0.0f,
			 1.0f,  1.0f, 0.0f, 1.0f, 1.0f
		};
		glGenVertexArrays(1, &quadVAO);
		glGenBuffers(1, &quadVBO);
		glBindVertexArray(quadVAO);
		glBindBuffer(GL_ARRAY_BUFFER, quadVBO);
		glBufferData(GL_ARRAY_BUFFER, sizeof(quadVertices), &quadVertices, GL_STATIC_DRAW);
		glEnableVertexAttribArray(0);
		glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 5 * sizeof(GLfloat), (GLvoid*)0);
		glEnableVertexAttribArray(1);
		glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, 5 * sizeof(GLfloat), (GLvoid*)(3 * sizeof(GLfloat)));
	}
	glBindVertexArray(quadVAO);
	glDrawArrays(GL_TRIANGLES, 0, 6);
	glBindVertexArray(0);
}

void renderRobots(bool animation, float time) {
	// Base translation offset for the robots
	glm::vec3 baseTranslation = glm::vec3(-9.0f, 0.0f, 12.0f);
	glm::vec3 translationOffset = glm::vec3(2.0f, 0.0f, 0.0f); // Offset between robots in a row
	glm::vec3 rowOffset = glm::vec3(0.0f, 0.0f, 3.0f); // Offset between rows

	// Scaling factor for the entire robot
	float scaleFactor = 0.25f;

	// time for robot animation
	float timeForAnimaton = time * (157 * 4 / 60); // 157BPM Music

	float torsoAngle = 10.0f;
	float legAngle = 30.0f;
	

	// Calculate rotation angles (common for all robots)
	float torsoRotationAngle = glm::radians(torsoAngle) + glm::sin(timeForAnimaton) * glm::radians(torsoAngle); // Oscillates between 0 and 30 degrees
	float headRotationAngle = glm::sin(timeForAnimaton) * glm::radians(20.0f); // Oscillates between -20 and 20 degrees
	float leftArmRotationAngle = glm::sin(timeForAnimaton) * glm::radians(45.0f); // Left arm oscillation
	float rightArmRotationAngle = glm::sin((timeForAnimaton)+glm::pi<float>()) * glm::radians(45.0f); // Right arm offset by one period
	float leftLegRotationAngle = -(glm::radians(legAngle) + glm::sin((timeForAnimaton)+glm::pi<float>()) * glm::radians(20.0f));
	float rightLegRotationAngle = -(glm::radians(legAngle) + glm::sin((timeForAnimaton)) * glm::radians(legAngle));
	
	
	float leftLowerLegRotationAngle = glm::radians(legAngle) + glm::sin((timeForAnimaton)+glm::pi<float>()) * glm::radians(20.0f);
	float rightLowerLegRotationAngle = glm::radians(legAngle) + glm::sin(timeForAnimaton) * glm::radians(legAngle);

	// Original pivot points
	glm::vec3 robotTPivot = glm::vec3(-0.000657f, 4.27716f, 0.310848f);
	glm::vec3 robotHPivot = glm::vec3(-0.000657f, 9.09354f, 0.300734f);
	glm::vec3 robotLAPivot = glm::vec3(2.1416f, 8.51987f, 0.293726f);
	glm::vec3 robotRAPivot = glm::vec3(-2.1416f, 8.51987f, 0.293726f);
	glm::vec3 robotLLPivot = glm::vec3(0.629473f, 4.27716f, 0.310848f);
	glm::vec3 robotRLPivot = glm::vec3(-0.630786f, 4.27716f, 0.310848f);
	glm::vec3 robotLLLPivot = glm::vec3(0.662548f, 2.69585f, 0.303483f);
	glm::vec3 robotRLLPivot = glm::vec3(-0.663861f, 2.69585f, 0.303483f);

	for (float row = 0; row < 8; ++row) { // Loop for rows
		for (float col = 0; col < 10; ++col) { // Loop for robots in a row
			// Calculate the translation for the current robot
			glm::vec3 robotTranslation = baseTranslation + col * translationOffset +  row * rowOffset;

			// Model matrix for torso (start of hierarchy)
			glm::mat4 torsoModel = glm::mat4(1.0f);
			torsoModel = glm::translate(torsoModel, robotTranslation);   // Translate to robot's base position
			torsoModel = glm::scale(torsoModel, glm::vec3(scaleFactor)); // Apply uniform scaling
			torsoModel = glm::rotate(torsoModel, glm::radians(180.0f), glm::vec3(0.0f, 1.0f, 0.0f)); // Rotate 180 degrees around Y-axis
			torsoModel = glm::translate(torsoModel, glm::vec3(0.0f, -0.4f, 0.0f));
			if (animation) {
				torsoModel = glm::translate(torsoModel, robotTPivot);
				torsoModel = glm::rotate(torsoModel, torsoRotationAngle, glm::vec3(1.0f, 0.0f, 0.0f));
				torsoModel = glm::translate(torsoModel, -robotTPivot);
			}
			// Model matrix for head
			glm::mat4 headModel = torsoModel;
			if (animation) {
				headModel = glm::translate(headModel, robotHPivot);
				headModel = glm::rotate(headModel, headRotationAngle, glm::vec3(0.5f, 1.0f, 0.0f));
				headModel = glm::translate(headModel, -robotHPivot);
			}
			// Model matrix for left arm
			glm::mat4 leftArmModel = torsoModel;
			if (animation) {
				leftArmModel = glm::translate(leftArmModel, robotLAPivot);
				leftArmModel = glm::rotate(leftArmModel, leftArmRotationAngle, glm::vec3(1.0f, 0.0f, 0.0f));
				leftArmModel = glm::translate(leftArmModel, -robotLAPivot);
			}

			// Model matrix for right arm
			glm::mat4 rightArmModel = torsoModel;
			if (animation) {
				rightArmModel = glm::translate(rightArmModel, robotRAPivot);
				rightArmModel = glm::rotate(rightArmModel, rightArmRotationAngle, glm::vec3(1.0f, 0.0f, 0.0f));
				rightArmModel = glm::translate(rightArmModel, -robotRAPivot);
			}




			// Model matrix for right leg
			glm::mat4 rightLegModel = glm::mat4(1.0f);
			rightLegModel = glm::translate(rightLegModel, robotTranslation);
			rightLegModel = glm::scale(rightLegModel, glm::vec3(scaleFactor));
			rightLegModel = glm::rotate(rightLegModel, glm::radians(180.0f), glm::vec3(0.0f, 1.0f, 0.0f));
			if (animation) {
				rightLegModel = glm::translate(rightLegModel, robotRLPivot);
				rightLegModel = glm::rotate(rightLegModel, rightLegRotationAngle, glm::vec3(1.0f, 0.0f, 0.0f));
				rightLegModel = glm::translate(rightLegModel, -robotRLPivot);
			}

			// Model matrix for left leg
			glm::mat4 leftLegModel = glm::mat4(1.0f);
			leftLegModel = glm::translate(leftLegModel, robotTranslation);
			leftLegModel = glm::scale(leftLegModel, glm::vec3(scaleFactor));
			leftLegModel = glm::rotate(leftLegModel, glm::radians(180.0f), glm::vec3(0.0f, 1.0f, 0.0f));
			if (animation) {
				leftLegModel = glm::translate(leftLegModel, robotLLPivot);
				leftLegModel = glm::rotate(leftLegModel, leftLegRotationAngle, glm::vec3(1.0f, 0.0f, 0.0f));
				leftLegModel = glm::translate(leftLegModel, -robotLLPivot);
			}

			// Model matrix for right lower leg
			glm::mat4 rightLowerLegModel = rightLegModel;
			if (animation) {
				rightLowerLegModel = glm::translate(rightLowerLegModel, robotRLLPivot);
				rightLowerLegModel = glm::rotate(rightLowerLegModel, rightLowerLegRotationAngle, glm::vec3(1.0f, 0.0f, 0.0f));
				rightLowerLegModel = glm::translate(rightLowerLegModel, -robotRLLPivot);
			}

			// Model matrix for left lower leg
			glm::mat4 leftLowerLegModel = leftLegModel;
			if (animation) {
				leftLowerLegModel = glm::translate(leftLowerLegModel, robotLLLPivot);
				leftLowerLegModel = glm::rotate(leftLowerLegModel, leftLowerLegRotationAngle, glm::vec3(1.0f, 0.0f, 0.0f));
				leftLowerLegModel = glm::translate(leftLowerLegModel, -robotLLLPivot);
			}
			// Render all Robot Parts
			glUniformMatrix4fv(glGetUniformLocation(geometryShader, "model"), 1, GL_FALSE, glm::value_ptr(torsoModel));
			robotT.RenderModel();
			glUniformMatrix4fv(glGetUniformLocation(geometryShader, "model"), 1, GL_FALSE, glm::value_ptr(headModel));
			robotH.RenderModel();
			glUniformMatrix4fv(glGetUniformLocation(geometryShader, "model"), 1, GL_FALSE, glm::value_ptr(leftArmModel));
			robotLA.RenderModel();
			glUniformMatrix4fv(glGetUniformLocation(geometryShader, "model"), 1, GL_FALSE, glm::value_ptr(rightArmModel));
			robotRA.RenderModel();
			glUniformMatrix4fv(glGetUniformLocation(geometryShader, "model"), 1, GL_FALSE, glm::value_ptr(rightLegModel));
			robotRL.RenderModel();
			glUniformMatrix4fv(glGetUniformLocation(geometryShader, "model"), 1, GL_FALSE, glm::value_ptr(leftLegModel));
			robotLL.RenderModel();
			glUniformMatrix4fv(glGetUniformLocation(geometryShader, "model"), 1, GL_FALSE, glm::value_ptr(rightLowerLegModel));
			robotRLL.RenderModel();
			glUniformMatrix4fv(glGetUniformLocation(geometryShader, "model"), 1, GL_FALSE, glm::value_ptr(leftLowerLegModel));
			robotLLL.RenderModel();
		}
	}
}
void renderDJ(bool animation, float time) {
	// Base translation offset for the robots
	glm::vec3 baseTranslation = glm::vec3(0.0f, 2.0f, 3.0f);

	// Scaling factor for the entire robot
	float scaleFactor = 0.35f;

	// time for robot animation
	float timeForAnimaton = time * (157 * 4 / 60); // 157BPM Music

	float torsoAngle = 10.0f;
	float legAngle = 30.0f;


	// Calculate rotation angles (common for all robots)

	float torsoRotationAngle = glm::radians(torsoAngle) + glm::sin(timeForAnimaton) * glm::radians(torsoAngle); // Oscillates between 0 and 30 degrees
	float headRotationAngle = glm::sin(timeForAnimaton) * glm::radians(20.0f); // Oscillates between -20 and 20 degrees
	float leftArmRotationAngle = glm::radians(180.0f) +  glm::sin(timeForAnimaton) * glm::radians(45.0f); // Left arm oscillation
	float rightArmRotationAngle = glm::sin((timeForAnimaton)+glm::pi<float>()) * glm::radians(5.0f); // Right arm offset by one period

	// Original pivot points
	glm::vec3 robotTPivot = glm::vec3(-0.000657f, 4.27716f, 0.310848f);
	glm::vec3 robotHPivot = glm::vec3(-0.000657f, 9.09354f, 0.300734f);
	glm::vec3 robotLAPivot = glm::vec3(2.1416f, 8.51987f, 0.293726f);
	glm::vec3 robotRAPivot = glm::vec3(-2.1416f, 8.51987f, 0.293726f);
	glm::vec3 robotLLPivot = glm::vec3(0.629473f, 4.27716f, 0.310848f);
	glm::vec3 robotRLPivot = glm::vec3(-0.630786f, 4.27716f, 0.310848f);

	glm::vec3 robotTranslation = baseTranslation;

	// Model matrix for torso (start of hierarchy)
	glm::mat4 torsoModel = glm::mat4(1.0f);
	torsoModel = glm::translate(torsoModel, robotTranslation);   // Translate to DJ's base position
	torsoModel = glm::scale(torsoModel, glm::vec3(scaleFactor)); // Apply uniform scaling
	torsoModel = glm::translate(torsoModel, glm::vec3(0.0f, -0.4f, 0.0f));
	if (animation) {
		torsoModel = glm::translate(torsoModel, robotTPivot);
		torsoModel = glm::rotate(torsoModel, torsoRotationAngle, glm::vec3(1.0f, 0.0f, 0.0f));
		torsoModel = glm::translate(torsoModel, -robotTPivot);
	}
	// Model matrix for head
	glm::mat4 headModel = torsoModel;
	if (animation) {
		headModel = glm::translate(headModel, robotHPivot);
		headModel = glm::rotate(headModel, headRotationAngle, glm::vec3(0.5f, 1.0f, 0.0f));
		headModel = glm::translate(headModel, -robotHPivot);
	}
	// Model matrix for left arm
	glm::mat4 leftArmModel = torsoModel;
	if (animation) {
		leftArmModel = glm::translate(leftArmModel, robotLAPivot);
		leftArmModel = glm::rotate(leftArmModel, leftArmRotationAngle, glm::vec3(1.0f, 0.0f, 0.0f));
		leftArmModel = glm::translate(leftArmModel, -robotLAPivot);
	}

	// Model matrix for right arm
	glm::mat4 rightArmModel = torsoModel;
	if (animation) {
		rightArmModel = glm::translate(rightArmModel, robotRAPivot);
		rightArmModel = glm::rotate(rightArmModel, rightArmRotationAngle, glm::vec3(1.0f, 0.0f, 0.0f));
		rightArmModel = glm::translate(rightArmModel, -robotRAPivot);
	}

	// Model matrix for right leg
	glm::mat4 rightLegModel = glm::mat4(1.0f);
	rightLegModel = glm::translate(rightLegModel, robotTranslation);
	rightLegModel = glm::scale(rightLegModel, glm::vec3(scaleFactor));

	// Model matrix for left leg
	glm::mat4 leftLegModel = glm::mat4(1.0f);
	leftLegModel = glm::translate(leftLegModel, robotTranslation);
	leftLegModel = glm::scale(leftLegModel, glm::vec3(scaleFactor));

	// Render all DJ Parts
	glUniformMatrix4fv(glGetUniformLocation(geometryShader, "model"), 1, GL_FALSE, glm::value_ptr(torsoModel));
	DJT.RenderModel();
	glUniformMatrix4fv(glGetUniformLocation(geometryShader, "model"), 1, GL_FALSE, glm::value_ptr(headModel));
	DJH.RenderModel();
	glUniformMatrix4fv(glGetUniformLocation(geometryShader, "model"), 1, GL_FALSE, glm::value_ptr(leftArmModel));
	DJLA.RenderModel();
	glUniformMatrix4fv(glGetUniformLocation(geometryShader, "model"), 1, GL_FALSE, glm::value_ptr(rightArmModel));
	DJRA.RenderModel();
	glUniformMatrix4fv(glGetUniformLocation(geometryShader, "model"), 1, GL_FALSE, glm::value_ptr(rightLegModel));
	DJRL.RenderModel();
	glUniformMatrix4fv(glGetUniformLocation(geometryShader, "model"), 1, GL_FALSE, glm::value_ptr(leftLegModel));
	DJLL.RenderModel();
}


glm::vec3 getRadians(glm::vec3 direction) {
	glm::vec3 defaultDirection = glm::vec3(0.0f, -1.0f, 0.0f);

	// Calculate angle in radians between the two vectors
	float angle = glm::orientedAngle(glm::normalize(direction), glm::normalize(defaultDirection), glm::vec3(0.0f, 0.0f, 1.0f));

	// Return the angle as a vec3 (you might need to adjust this based on how you use it)
	return glm::vec3(angle, 0.0f, 0.0f);
}

// Function for geometry pass
void geometryPass(bool robotAnim, float time) {
	glBindFramebuffer(GL_FRAMEBUFFER, gBuffer);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glUseProgram(geometryShader);

	// Camera Uniforms
	glm::mat4 view = camera.calculateViewMatrix();
	projection = glm::perspective(glm::radians(45.0f), (GLfloat)mainWindow.getBufferWidth() / mainWindow.getBufferHeight(), nearZ, farZ);
	glUniformMatrix4fv(glGetUniformLocation(geometryShader, "view"), 1, GL_FALSE, glm::value_ptr(view));
	glUniformMatrix4fv(glGetUniformLocation(geometryShader, "projection"), 1, GL_FALSE, glm::value_ptr(projection));

	glm::mat4 model(1.0f);

	// Render stage
	glUniformMatrix4fv(glGetUniformLocation(geometryShader, "model"), 1, GL_FALSE, glm::value_ptr(model));
	stage.RenderModel();

	// Render lamp
	for (int i = 0; i < MAX_SPOT_LIGHTS; i++) {
		glm::mat4 lampModel(1.0f);
		glm::vec3 lampTranslation = concertLights[i].getPosition();
		glm::vec3 lampRotation = getRadians(concertLights[i].getRotation()); // Get the rotation from ConcertLight
		float lampScale = 2.5f;

		// lampModel = glm::translate(lampModel, lampTranslation);
		// lampModel = glm::rotate(lampModel, glm::radians(90.0f), glm::vec3(1.0f, 0.0f, 0.0f));
		
		// 1. Create the rotation matrices (same as before)
		glm::mat4 rotateX = glm::rotate(glm::mat4(1.0f), glm::radians(lampRotation.x), glm::vec3(1.0f, 0.0f, 0.0f));
		glm::mat4 rotateY = glm::rotate(glm::mat4(1.0f), glm::radians(lampRotation.y), glm::vec3(0.0f, 1.0f, 0.0f));
		glm::mat4 rotateZ = glm::rotate(glm::mat4(1.0f), glm::radians(lampRotation.z), glm::vec3(0.0f, 0.0f, 1.0f));

		// 2. Apply rotations (before translation)
		lampModel = rotateZ * rotateY * rotateX * lampModel;

		// 3. Translate to the correct position (same as before)
		lampModel = glm::translate(lampModel, lampTranslation);
		lampModel = glm::rotate(lampModel, glm::radians(90.0f), glm::vec3(1.0f, 0.0f, 0.0f));
		
		lampModel = glm::scale(lampModel, glm::vec3(lampScale));

		glUniformMatrix4fv(glGetUniformLocation(geometryShader, "model"), 1, GL_FALSE, glm::value_ptr(lampModel));
		concertLights[i].getLamp().RenderModel();

		lampModel = glm::translate(lampModel, -lampTranslation);
		glUniformMatrix4fv(glGetUniformLocation(geometryShader, "model"), 1, GL_FALSE, glm::value_ptr(lampModel));
	}



	renderRobots(robotAnim, time);

	renderDJ(robotAnim, time);

	
	glm::mat4 mixtableModel(1.0f);
	mixtableModel = glm::translate(mixtableModel, glm::vec3(0.0f,2.1f,4.3f));
	mixtableModel = glm::scale(mixtableModel, glm::vec3(0.3f));
	mixtableModel = glm::rotate(mixtableModel, glm::radians(180.0f), glm::vec3(0.0f, 1.0f, 0.0f));
	glUniformMatrix4fv(glGetUniformLocation(geometryShader, "model"), 1, GL_FALSE, glm::value_ptr(mixtableModel));
	Mixtable.RenderModel();
	

	// Render plane
	glUniformMatrix4fv(glGetUniformLocation(geometryShader, "model"), 1, GL_FALSE, glm::value_ptr(model));
	terrain.RenderModel();
	model = glm::scale(model, glm::vec3(1.0f / 30.0f, 1.0f, 1.0f / 30.0f));
	/*
	// Render tree
	glUniformMatrix4fv(glGetUniformLocation(geometryShader, "model"), 1, GL_FALSE, glm::value_ptr(model));
	tree.RenderModel();
	*/

	glBindFramebuffer(GL_FRAMEBUFFER, 0);
}

// Function for lighting pass
void lightingPass() {
	glBindFramebuffer(GL_FRAMEBUFFER, lightingFramebuffer);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glUseProgram(lightingShader);

	// Bind G-buffer textures
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, gPosition);
	glUniform1i(glGetUniformLocation(lightingShader, "gPosition"), 0);
	glActiveTexture(GL_TEXTURE1);
	glBindTexture(GL_TEXTURE_2D, gNormal);
	glUniform1i(glGetUniformLocation(lightingShader, "gNormal"), 1);
	glActiveTexture(GL_TEXTURE2);
	glBindTexture(GL_TEXTURE_2D, gAlbedoSpec);
	glUniform1i(glGetUniformLocation(lightingShader, "gAlbedoSpec"), 2);

	glm::mat4 view = camera.calculateViewMatrix();

	// Set pointlight uniforms
	for (unsigned int i = 0; i < MAX_POINT_LIGHTS; ++i) {
		std::string number = std::to_string(i);
		std::string posUniformName = "pointLights[" + number + "].Position";
		std::string colUniformName = "pointLights[" + number + "].Color";
		std::string ambientIntensityUniformName = "pointLights[" + number + "].ambientIntensity";
		std::string diffuseIntensityUniformName = "pointLights[" + number + "].diffuseIntensity";



		// Transform light position to view space
		glm::vec3 lightPosView = glm::vec3(view * glm::vec4(pointLights[i].getPosition(), 1.0));
		glUniform3fv(glGetUniformLocation(lightingShader, posUniformName.c_str()), 1, glm::value_ptr(lightPosView));
		glUniform3fv(glGetUniformLocation(lightingShader, colUniformName.c_str()), 1, glm::value_ptr(pointLights[i].getColor()));
		glUniform1f(glGetUniformLocation(lightingShader, ambientIntensityUniformName.c_str()), pointLights[i].getAmbientIntensity());
		glUniform1f(glGetUniformLocation(lightingShader, diffuseIntensityUniformName.c_str()), pointLights[i].getDiffuseIntensity());
	}
	glUniform1i(glGetUniformLocation(lightingShader, "numPointLights"), MAX_POINT_LIGHTS);

	// Set spotlight uniforms
	for (unsigned int i = 0; i < MAX_SPOT_LIGHTS; ++i) {
		std::string number = std::to_string(i);
		std::string posUniformName = "spotLights[" + number + "].Position";
		std::string colUniformName = "spotLights[" + number + "].Color";
		std::string dirUniformName = "spotLights[" + number + "].Direction";
		std::string ambientIntensityUniformName = "spotLights[" + number + "].ambientIntensity";
		std::string diffuseIntensityUniformName = "spotLights[" + number + "].diffuseIntensity";
		std::string cutOffUniformName = "spotLights[" + number + "].CutOff";
		std::string outerCutOffUniformName = "spotLights[" + number + "].OuterCutOff";

		// Calculate the new position with the wiggle offset (call updateLights() before this)
		glm::vec3 newPosition = concertLights[i].getPosition();

		// Transform light position to view space
		glm::vec3 lightPosView = glm::vec3(view * glm::vec4(newPosition, 1.0));
		// Transform light direction to view space
		glm::vec3 lightDirView = glm::normalize(glm::mat3(view) * concertLights[i].getRotation());
		glUniform3fv(glGetUniformLocation(lightingShader, posUniformName.c_str()), 1, glm::value_ptr(lightPosView));
		glUniform3fv(glGetUniformLocation(lightingShader, dirUniformName.c_str()), 1, glm::value_ptr(lightDirView));
		glUniform3fv(glGetUniformLocation(lightingShader, colUniformName.c_str()), 1, glm::value_ptr(concertLights[i].getSpotlight().getColor()));
		glUniform1f(glGetUniformLocation(lightingShader, ambientIntensityUniformName.c_str()), concertLights[i].getSpotlight().getAmbientIntensity());
		glUniform1f(glGetUniformLocation(lightingShader, diffuseIntensityUniformName.c_str()), concertLights[i].getSpotlight().getDiffuseIntensity());
		glUniform1f(glGetUniformLocation(lightingShader, cutOffUniformName.c_str()), cos(glm::radians(concertLights[i].getSpotlight().getEdge())));
		glUniform1f(glGetUniformLocation(lightingShader, outerCutOffUniformName.c_str()), cos(glm::radians(concertLights[i].getSpotlight().getEdge() + 5.0f))); // Outer cutoff slightly larger
	}
	glUniform1i(glGetUniformLocation(lightingShader, "numSpotLights"), MAX_SPOT_LIGHTS);


	// Transform directional light direction to view space
	glm::vec3 directionalDirView = glm::normalize(glm::vec3(view * glm::vec4(mainDirectionalLight.getDirection(), 0.0)));
	glUniform3fv(glGetUniformLocation(lightingShader, "directionalLight.direction"), 1, glm::value_ptr(directionalDirView));
	glUniform3fv(glGetUniformLocation(lightingShader, "directionalLight.color"), 1, glm::value_ptr(mainDirectionalLight.getColor()));
	glUniform1f(glGetUniformLocation(lightingShader, "directionalLight.ambientIntensity"), mainDirectionalLight.getAmbientIntensity());
	glUniform1f(glGetUniformLocation(lightingShader, "directionalLight.diffuseIntensity"), mainDirectionalLight.getDiffuseIntensity());

	// Pass view position in view space (which is typically the origin)
	glUniform3fv(glGetUniformLocation(lightingShader, "viewPos"), 1, glm::value_ptr(glm::vec3(0.0, 0.0, 0.0)));

	renderQuad();
	glBindFramebuffer(GL_FRAMEBUFFER, 0);
}

// Function for combining pass
void combinePass(bool useSSAO) {
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glUseProgram(combineShader);

	// Bind lighting texture to texture unit 0
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, lightingTexture);
	glUniform1i(glGetUniformLocation(combineShader, "lightingTexture"), 0);

	// Bind albedo texture to texture unit 1
	glActiveTexture(GL_TEXTURE1);
	glBindTexture(GL_TEXTURE_2D, gAlbedoSpec);
	glUniform1i(glGetUniformLocation(combineShader, "gAlbedoSpec"), 1);

	// Bind SSAO texture to texture unit 2
	glActiveTexture(GL_TEXTURE2);
	glBindTexture(GL_TEXTURE_2D, ssaoColorBufferBlur);
	glUniform1i(glGetUniformLocation(combineShader, "ssao"), 2);

	// Pass the SSAO usage flag to the shader
	glUniform1i(glGetUniformLocation(combineShader, "useSSAO"), useSSAO);

	renderQuad();
}

// Function to set up SSAO
void setupSSAO() {
	// Generate SSAO framebuffer
	glGenFramebuffers(1, &ssaoFBO);
	glBindFramebuffer(GL_FRAMEBUFFER, ssaoFBO);

	// SSAO color buffer
	glGenTextures(1, &ssaoColorBuffer);
	glBindTexture(GL_TEXTURE_2D, ssaoColorBuffer);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RED, mainWindow.getBufferWidth(), mainWindow.getBufferHeight(), 0, GL_RED, GL_FLOAT, NULL);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, ssaoColorBuffer, 0);

	if (glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE) {
		std::cout << "SSAO Framebuffer not complete!" << std::endl;
	}

	// Generate Blur framebuffer
	glGenFramebuffers(1, &ssaoBlurFBO);
	glBindFramebuffer(GL_FRAMEBUFFER, ssaoBlurFBO);

	// SSAO blur color buffer
	glGenTextures(1, &ssaoColorBufferBlur);
	glBindTexture(GL_TEXTURE_2D, ssaoColorBufferBlur);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RED, mainWindow.getBufferWidth(), mainWindow.getBufferHeight(), 0, GL_RED, GL_FLOAT, NULL);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, ssaoColorBufferBlur, 0);

	if (glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE) {
		std::cout << "SSAO Blur Framebuffer not complete!" << std::endl;
	}

	glBindFramebuffer(GL_FRAMEBUFFER, 0);

	// SSAO Kernel
	std::uniform_real_distribution<GLfloat> randomFloats(0.0, 1.0); // random floats between 0.0 - 1.0
	std::default_random_engine generator;
	ssaoKernel.clear();

	for (unsigned int i = 0; i < 64; ++i) {
		glm::vec3 sample(
			randomFloats(generator) * 2.0 - 1.0, // x
			randomFloats(generator) * 2.0 - 1.0, // y
			randomFloats(generator)              // z
		);
		sample = glm::normalize(sample);
		sample *= randomFloats(generator);
		float scale = (float)i / 64.0;
		scale = lerp(0.1f, 1.0f, scale * scale);
		sample *= scale;
		ssaoKernel.push_back(sample);
	}

	// SSAO Noise Texture
	std::vector<glm::vec3> ssaoNoise;
	for (unsigned int i = 0; i < 16; i++) {
		glm::vec3 noise(
			randomFloats(generator) * 2.0 - 1.0,
			randomFloats(generator) * 2.0 - 1.0,
			0.0f); // Rotate around z-axis
		ssaoNoise.push_back(noise);
	}

	glGenTextures(1, &noiseTexture);
	glBindTexture(GL_TEXTURE_2D, noiseTexture);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F, 4, 4, 0, GL_RGB, GL_FLOAT, &ssaoNoise[0]);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
}


// Function for SSAO pass
void ssaoPass() {
	// Generate SSAO texture
	glBindFramebuffer(GL_FRAMEBUFFER, ssaoFBO);
	glClear(GL_COLOR_BUFFER_BIT);
	glUseProgram(ssaoShader);

	// Set uniforms
	for (unsigned int i = 0; i < 64; ++i) {
		glUniform3fv(glGetUniformLocation(ssaoShader, ("samples[" + std::to_string(i) + "]").c_str()), 1, &ssaoKernel[i][0]);
	}
	glm::mat4 view = camera.calculateViewMatrix();
	glUniform1i(glGetUniformLocation(ssaoShader, "gPosition"), 0);
	glUniform1i(glGetUniformLocation(ssaoShader, "gNormal"), 1);
	glUniform1i(glGetUniformLocation(ssaoShader, "noiseTexture"), 2);
	glUniformMatrix4fv(glGetUniformLocation(ssaoShader, "projection"), 1, GL_FALSE, glm::value_ptr(projection));
	glUniform2f(glGetUniformLocation(ssaoShader, "screenSize"), mainWindow.getBufferWidth(), mainWindow.getBufferHeight());
	glUniformMatrix4fv(glGetUniformLocation(ssaoShader, "view"), 1, GL_FALSE, glm::value_ptr(view));

	// Bind G-buffer textures
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, gPosition);
	glActiveTexture(GL_TEXTURE1);
	glBindTexture(GL_TEXTURE_2D, gNormal);
	glActiveTexture(GL_TEXTURE2);
	glBindTexture(GL_TEXTURE_2D, noiseTexture);

	renderQuad();
	glBindFramebuffer(GL_FRAMEBUFFER, 0);

	// 2. Blur SSAO texture to remove noise
	glBindFramebuffer(GL_FRAMEBUFFER, ssaoBlurFBO);
	glClear(GL_COLOR_BUFFER_BIT);
	glUseProgram(ssaoBlurShader);
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, ssaoColorBuffer);
	glUniform1i(glGetUniformLocation(ssaoBlurShader, "ssaoInput"), 0);
	renderQuad();
	glBindFramebuffer(GL_FRAMEBUFFER, 0);
}

// Function to display texture
void displayTexture(GLuint texture, bool isDepth = false) {
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glUseProgram(displayShader);
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, texture);
	glUniform1i(glGetUniformLocation(displayShader, "displayTexture"), 0);
	glUniform1i(glGetUniformLocation(displayShader, "isDepth"), isDepth);
	renderQuad();
}


void initializeParticles() {
	// Parameters for all particle systems:
	int particleMax = 1000000;
	float TTL = 0.2f;
	float particleSize = 0.2f;
	int particlesPerSecond = 10000;

	// Circular Particle System
	circularParticleSystemStage = new ParticleSystem(
		particleMax, TTL, particleSize, particlesPerSecond, orangeColor , nearZ, farZ
	);
	circularParticleSystemStage2 = new ParticleSystem(
		particleMax, TTL, particleSize, particlesPerSecond, purpleColor, nearZ, farZ
	);
	ParticleCircularMovement* circularMovement = new ParticleCircularMovement(
		glm::vec3(0.0f, 7.5f, 0.0f), 4.0f, 2.0f, 0.0f
	);
	ParticleCircularMovement* circularMovement2 = new ParticleCircularMovement(
		glm::vec3(0.0f, 7.5f, 0.0f), 4.0f, 2.0f, PI
	);
	circularParticleSystemStage->SetMovementPattern(circularMovement);
	circularParticleSystemStage2->SetMovementPattern(circularMovement2);

	// Linear Particle System
	float TTLLinear = 0.1f;
	float particleSizeLinear = 0.5f;
	float duration = (157.0f / 120.0f);
	int particlesPerSecondLinear = 10000;
	linearParticleSystem = new ParticleSystem(
		particleMax, TTLLinear, particleSizeLinear, particlesPerSecondLinear, purpleColor, nearZ, farZ
	);
	linearParticleSystem2 = new ParticleSystem(
		particleMax, TTLLinear, particleSizeLinear, particlesPerSecondLinear, purpleColor, nearZ, farZ
	);
	linearParticleSystem3 = new ParticleSystem(
		particleMax, TTLLinear, particleSizeLinear, particlesPerSecondLinear, orangeColor, nearZ, farZ
	);
	linearParticleSystem4 = new ParticleSystem(
		particleMax, TTLLinear, particleSizeLinear, particlesPerSecondLinear, orangeColor, nearZ, farZ
	);
	linearParticleSystem5 = new ParticleSystem(
		particleMax, TTLLinear, particleSizeLinear, particlesPerSecondLinear, purpleColor, nearZ, farZ
	);
	linearParticleSystem6 = new ParticleSystem(
		particleMax, TTLLinear, particleSizeLinear, particlesPerSecondLinear, purpleColor, nearZ, farZ
	);
	linearParticleSystem7 = new ParticleSystem(
		particleMax, TTLLinear, particleSizeLinear, particlesPerSecondLinear, orangeColor, nearZ, farZ
	);
	linearParticleSystem8 = new ParticleSystem(
		particleMax, TTLLinear, particleSizeLinear, particlesPerSecondLinear, orangeColor, nearZ, farZ
	);
	linearParticleSystem9 = new ParticleSystem(
		particleMax, TTLLinear, particleSizeLinear, particlesPerSecondLinear, purpleColor, nearZ, farZ
	);
	linearParticleSystem10 = new ParticleSystem(
		particleMax, TTLLinear, particleSizeLinear, particlesPerSecondLinear, orangeColor, nearZ, farZ
	);
	linearParticleSystem11 = new ParticleSystem(
		particleMax, TTLLinear, particleSizeLinear, particlesPerSecondLinear, orangeColor, nearZ, farZ
	);
	linearParticleSystem12 = new ParticleSystem(
		particleMax, TTLLinear, particleSizeLinear, particlesPerSecondLinear, purpleColor, nearZ, farZ
	);
	float bottom = 3.0f;
	float side = 12.0f;
	float front = 9.0f;
	float back = 35.0f;
	float yOffset = 3.0f;
	float xOffset = 2.0f * side / 5.0f;
	ParticleLinearMovement* linearMovement = new ParticleLinearMovement(
		glm::vec3(side, bottom, back),
		glm::vec3(side, bottom, front),
		duration
	);
	ParticleLinearMovement* linearMovement2 = new ParticleLinearMovement(
		glm::vec3(-side, bottom, back),
		glm::vec3(-side, bottom, front),
		duration
	);
	ParticleLinearMovement* linearMovement3 = new ParticleLinearMovement(
		glm::vec3(side, bottom + yOffset, back),
		glm::vec3(side, bottom + yOffset, front),
		duration
	);
	ParticleLinearMovement* linearMovement4 = new ParticleLinearMovement(
		glm::vec3(-side, bottom + yOffset, back),
		glm::vec3(-side, bottom + yOffset, front),
		duration
	);
	ParticleLinearMovement* linearMovement5 = new ParticleLinearMovement(
		glm::vec3(side, bottom + 2 * yOffset, back),
		glm::vec3(side, bottom + 2 * yOffset, front),
		duration
	);
	ParticleLinearMovement* linearMovement6 = new ParticleLinearMovement(
		glm::vec3(-side, bottom + 2 * yOffset, back),
		glm::vec3(-side, bottom + 2 * yOffset, front),
		duration
	);
	ParticleLinearMovement* linearMovement7 = new ParticleLinearMovement(
		glm::vec3(side, bottom + 3 * yOffset, back),
		glm::vec3(side, bottom + 3 * yOffset, front),
		duration
	);
	ParticleLinearMovement* linearMovement8 = new ParticleLinearMovement(
		glm::vec3(-side, bottom + 3 * yOffset, back),
		glm::vec3(-side, bottom + 3 * yOffset, front),
		duration
	);
	ParticleLinearMovement* linearMovement9 = new ParticleLinearMovement(
		glm::vec3(-side + xOffset, bottom + 3 * yOffset, back),
		glm::vec3(-side + xOffset, bottom + 3 * yOffset, front),
		duration
	);
	ParticleLinearMovement* linearMovement10 = new ParticleLinearMovement(
		glm::vec3(-side + 2 * xOffset, bottom + 3 * yOffset, back),
		glm::vec3(-side + 2 * xOffset, bottom + 3 * yOffset, front),
		duration
	);
	ParticleLinearMovement* linearMovement11 = new ParticleLinearMovement(
		glm::vec3(-side + 3 * xOffset, bottom + 3 * yOffset, back),
		glm::vec3(-side + 3 * xOffset, bottom + 3 * yOffset, front),
		duration
	);
	ParticleLinearMovement* linearMovement12 = new ParticleLinearMovement(
		glm::vec3(-side + 4 * xOffset, bottom + 3 * yOffset, back),
		glm::vec3(-side + 4 * xOffset, bottom + 3 * yOffset, front),
		duration
	);




	linearMovement->Start(glfwGetTime());
	linearMovement2->Start(glfwGetTime());
	linearMovement3->Start(glfwGetTime() + duration / 2);
	linearMovement4->Start(glfwGetTime() + duration / 2);
	linearMovement5->Start(glfwGetTime());
	linearMovement6->Start(glfwGetTime());
	linearMovement7->Start(glfwGetTime() + duration / 2);
	linearMovement8->Start(glfwGetTime() + duration / 2);
	linearMovement9->Start(glfwGetTime());
	linearMovement10->Start(glfwGetTime() + duration / 2);
	linearMovement11->Start(glfwGetTime() + duration / 2);
	linearMovement12->Start(glfwGetTime());
	linearParticleSystem->SetMovementPattern(linearMovement);
	linearParticleSystem2->SetMovementPattern(linearMovement2);
	linearParticleSystem3->SetMovementPattern(linearMovement3);
	linearParticleSystem4->SetMovementPattern(linearMovement4);
	linearParticleSystem5->SetMovementPattern(linearMovement5);
	linearParticleSystem6->SetMovementPattern(linearMovement6);
	linearParticleSystem7->SetMovementPattern(linearMovement7);
	linearParticleSystem8->SetMovementPattern(linearMovement8);
	linearParticleSystem9->SetMovementPattern(linearMovement9);
	linearParticleSystem10->SetMovementPattern(linearMovement10);
	linearParticleSystem11->SetMovementPattern(linearMovement11);
	linearParticleSystem12->SetMovementPattern(linearMovement12);





	circularParticleSystemStage->Initialize();
	circularParticleSystemStage2->Initialize();
	linearParticleSystem->Initialize();
	linearParticleSystem2->Initialize();
	linearParticleSystem3->Initialize();
	linearParticleSystem4->Initialize();
	linearParticleSystem5->Initialize();
	linearParticleSystem6->Initialize();
	linearParticleSystem7->Initialize();
	linearParticleSystem8->Initialize();
	linearParticleSystem9->Initialize();
	linearParticleSystem10->Initialize();
	linearParticleSystem11->Initialize();
	linearParticleSystem12->Initialize();
}



// Function to update particles
void updateParticles(float deltaTime, bool kickDrop) {
	circularParticleSystemStage->Update(deltaTime);
	circularParticleSystemStage2->Update(deltaTime);
	if (kickDrop) {
		linearParticleSystem->Update(deltaTime);
		linearParticleSystem2->Update(deltaTime);
		linearParticleSystem3->Update(deltaTime);
		linearParticleSystem4->Update(deltaTime);
		linearParticleSystem5->Update(deltaTime);
		linearParticleSystem6->Update(deltaTime);
		linearParticleSystem7->Update(deltaTime);
		linearParticleSystem8->Update(deltaTime);
		linearParticleSystem9->Update(deltaTime);
		linearParticleSystem10->Update(deltaTime);
		linearParticleSystem11->Update(deltaTime);
		linearParticleSystem12->Update(deltaTime);
	}
}

// Function to render particles
void renderParticles(bool kickDrop) {
	glm::vec3 camPos = camera.getCameraPosition();
	glm::mat4 camView = camera.calculateViewMatrix();
	glDepthFunc(GL_LEQUAL);
	glDisable(GL_DEPTH_TEST);
	glm::vec2 screenDimensions(mainWindow.getBufferWidth(), mainWindow.getBufferHeight());
	circularParticleSystemStage->Render(camView, projection, camPos, depthTexture, screenDimensions);
	circularParticleSystemStage2->Render(camView, projection, camPos, depthTexture, screenDimensions);
	if (kickDrop) {
		linearParticleSystem->Render(camView, projection, camPos, depthTexture, screenDimensions);
		linearParticleSystem2->Render(camView, projection, camPos, depthTexture, screenDimensions);
		linearParticleSystem3->Render(camView, projection, camPos, depthTexture, screenDimensions);
		linearParticleSystem4->Render(camView, projection, camPos, depthTexture, screenDimensions);
		linearParticleSystem5->Render(camView, projection, camPos, depthTexture, screenDimensions);
		linearParticleSystem6->Render(camView, projection, camPos, depthTexture, screenDimensions);
		linearParticleSystem7->Render(camView, projection, camPos, depthTexture, screenDimensions);
		linearParticleSystem8->Render(camView, projection, camPos, depthTexture, screenDimensions);
		linearParticleSystem9->Render(camView, projection, camPos, depthTexture, screenDimensions);
		linearParticleSystem10->Render(camView, projection, camPos, depthTexture, screenDimensions);
		linearParticleSystem11->Render(camView, projection, camPos, depthTexture, screenDimensions);
		linearParticleSystem12->Render(camView, projection, camPos, depthTexture, screenDimensions);
	}
	glEnable(GL_DEPTH_TEST);
}

void updateConcertLights(GLfloat deltaTime) {
	static float zOffset = 0.0f;
	static float direction = 1.0f; // 1.0f for moving forward, -1.0f for moving backward
	zOffset += 0.001f * direction;

	// Reverse direction when reaching a certain limit
	if (zOffset > 5.0f) { 
		direction = -1.0f;
	}
	else if (zOffset < -5.0f) {
		direction = 1.0f;
	}

	glm::vec3 red = glm::vec3(1.0f, 0.0f, 0.0f);
	glm::vec3 orange = glm::vec3(1.0f, 0.5f, 0.0f);
	glm::vec3 green = glm::vec3(0.0f, 1.0f, 0.0f);
	glm::vec3 blue = glm::vec3(0.0f, 0.0f, 1.0f);
	glm::vec3 purple = glm::vec3(0.5f, 0.0f, 1.0f);
	glm::vec3 yellow = glm::vec3(1.0f, 1.0f, 0.0f);
	
	GLfloat beats8 = 3.06;

	for (int i = 0; i < MAX_SPOT_LIGHTS; i++) {
		GLfloat now = glfwGetTime() - demoStartTime;

		// concertLights[i].lookDJ();
		
		// Intro I
		if (now <= 4*beats8) {
			concertLights[i].colorChange(orange);
		}
		// Intro II
		else if (now > 4*beats8 && now <= 8*beats8) {
			concertLights[i].colorChange(purple);
			
		}
		// Beat I
		else if (now > 8*beats8 && now <= 16*beats8) {
			concertLights[i].colorChange(purple);
			if (!swirling) {
				concertLights[i].swirlButFaster(deltaTime);
				swirling = false;
			}
		}
		// Beat II & III
		else if (now > 16*beats8 && now <= 24*beats8) {
			concertLights[i].lookStage();
			concertLights[i].colorChange(orange);
			concertLights[i].blink(true, deltaTime);
		}
		// Beat IV
		else if (now > 24 * beats8 && now <= 32 * beats8) {
			concertLights[i].lookDJ();
			concertLights[i].colorChange(purple);
			concertLights[i].blink(true, deltaTime);
			if (!swirling) {
				concertLights[i].swirlButFaster(deltaTime);
				swirling = false;
			}
		}
		// Beat V
		else if (now > 32 * beats8 && now <= 40 * beats8) {
			concertLights[i].lookStage();
			concertLights[i].colorChange(orange);
			if (!swirling) {
				concertLights[i].swirl();
				swirling = false;
			}
		}
		// Beat VI & VII
		else if (now > 40 * beats8 && now <= 48 * beats8) {
			concertLights[i].lookStage();
			concertLights[i].colorChange(purple);
			concertLights[i].blink(true, deltaTime);
		}
		// Break
		else if (now > 48 * beats8 && now <= 49 * beats8) {
			concertLights[i].lookStage();
			concertLights[i].colorChange(blue);
			concertLights[i].blink(false, deltaTime);
		}
		// Beat VIII
		else if (now > 49 * beats8 && now <= 54 * beats8) {
			concertLights[i].lookDJ();
			concertLights[i].colorChange(orange);
			concertLights[i].blink(true, deltaTime);
		}
		// Kinda Break
		else if (now > 54 * beats8 && now <= 56 * beats8) {
			concertLights[i].lookDJ();
			concertLights[i].colorChange(blue);
		}
		// Outro
		else if (now > 56 * beats8 && now <= 64 * beats8) {
			concertLights[i].lookDown();
			concertLights[i].colorChange(purple);
			concertLights[i].blink(false, deltaTime);
		}
	}
}

void InitializeBuffers() {
	setupGBuffer();
	setupLightingFramebuffer();
	setupSSAO();
}


int main() 
{
	// 
	ISoundEngine* soundEngine = createIrrKlangDevice();
	ISound* playingSound = soundEngine->play2D("Media/Nono_Interlude.wav", false, true, true);
	
	// Set up basic camera
	camera = Camera(glm::vec3(0.0f, 3.0f, 40.0f), glm::vec3(0.0f, 1.0f, 0.0f), 270.0f, 0.0f, 5.0f, 0.05f);

	// Set up DemoController
	DemoController demoController(camera, playingSound);

	// Initialize Window
	mainWindow = Window(1920, 1080);
	mainWindow.Initialise();

	// Initialize Buffers
	InitializeBuffers();

	// Initialize Shaders
	InitializeShaders();

	

	// Load Models
	LoadModels();

	// Initialize Lights
	initializeLights();

	// Initialize Particles
	initializeParticles();
	
	//print OpenGL Version
	const GLubyte* version = glGetString(GL_VERSION);
	std::cout << "OpenGL version: " << version << std::endl;

	// Set time variables
	GLfloat lastFrame = 0.0f;
	int frameCount = 0;
	int fps = 0;
	int avgFps = 0;
	float elapsedTime = 0.0f;
	// demoStartTime = 10000.0f;

	// Camera & Demo variables
	static bool f1KeyPressed = false;
	static bool mKeyPressed = false;

	demoController.startDemo(camera);
	demoStartTime = glfwGetTime();
	
	// Loop until window closed
	while (!mainWindow.getShouldClose())
	{
		// Time Calculation: 
		GLfloat now = glfwGetTime();
		deltaTime = now - lastFrame;
		lastFrame = now;
		elapsedTime += deltaTime;
		frameCount++;
		if (elapsedTime >= 1.0f)
		{
			fps = frameCount;
			frameCount = 0;
			elapsedTime = 0.0f;
			printf("\rCurrent FPS: %d", fps);
		}
		avgFps = (avgFps + fps) / 2;

		// Get + Handle User Input
		glfwPollEvents();

		// Check if we are inside the Demo Mode
		if (mainWindow.getKeys()[GLFW_KEY_F1]) {
			if (!f1KeyPressed) {
				if (!demoController.isDemoMode()) {
					demoController.startDemo(camera);
					demoStartTime = glfwGetTime();
				}
				else {
					demoController.endDemo();
				}
				f1KeyPressed = true;
			}
		}
		else {
			f1KeyPressed = false;
		}
		// Check if we press M for manual camera during demo
		if (mainWindow.getKeys()[GLFW_KEY_M]) {
			if (!mKeyPressed) {
				mKeyPressed = true;
			}		
			else {
			mKeyPressed = false;
			}
		}

		// Handle camera movement

		if (!demoController.isDemoMode() || mKeyPressed)
		{
			camera.keyControl(mainWindow.getKeys(), deltaTime);
			camera.mouseControl(mainWindow.getXChange(), mainWindow.getYChange());
			// Press K to get a new camera position for the Control Points
			if (mainWindow.getKeys()[GLFW_KEY_K]) { // Check if 'K' key is pressed
				glm::vec3 cameraPosition = camera.getCameraPosition();
				std::cout << "Camera Position: ("
					<< cameraPosition.x << ", "
					<< cameraPosition.y << ", "
					<< cameraPosition.z << ")" << std::endl;
			}
		}
		else 
		{
			demoController.moveCamera(deltaTime);
			
		}
		// Deferred Rendering
		float demoTimeElapsed = glfwGetTime() - demoStartTime;
		bool kickDrop = false;
		if (demoTimeElapsed >= 24.42f && demoTimeElapsed <= 185.0f) { // start robot animation/particle effects when kick drum comes
			kickDrop = true;
		}
		// Update Lamps and Spotlights
		updateConcertLights(deltaTime);
		updateParticles(deltaTime, kickDrop);

		// Clear the window
		glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		// Calculate projection matrix
		projection = glm::perspective(glm::radians(45.0f), (GLfloat)mainWindow.getBufferWidth() / mainWindow.getBufferHeight(), nearZ, farZ);
		geometryPass(kickDrop, demoTimeElapsed - 24.42f);
		ssaoPass();
		lightingPass();

		//Render Mode
		if (mainWindow.getKeys()[GLFW_KEY_1]) displayMode = 1; // Display position
		if (mainWindow.getKeys()[GLFW_KEY_2]) displayMode = 2; // Display normal
		if (mainWindow.getKeys()[GLFW_KEY_3]) displayMode = 3; // Display albedo/spec
		if (mainWindow.getKeys()[GLFW_KEY_4]) displayMode = 4; // Display lighting pass
		if (mainWindow.getKeys()[GLFW_KEY_5]) displayMode = 5; // Display depth Texture
		if (mainWindow.getKeys()[GLFW_KEY_6]) displayMode = 6; // Display combined rendering without particles
		if (mainWindow.getKeys()[GLFW_KEY_7]) displayMode = 7; // Display combined rendering with particles
		switch (displayMode) {
		case 1: displayTexture(gPosition); break;
		case 2: displayTexture(gNormal); break;
		case 3: displayTexture(gAlbedoSpec); break;
		case 4: displayTexture(lightingTexture); break;
		case 5: displayTexture(depthTexture, true); break;
		case 6: combinePass(false); break;
		default: 
			combinePass(false); 
			renderParticles(kickDrop); 
			break;
		}

		

		glUseProgram(0);

		mainWindow.swapBuffers();
	}

	soundEngine->drop();
	printf("\nAverage FPS: %d", avgFps);
	return 0;
}